/**
* @file math_library.h
 *
 * Utility classes and functions to simplify working with the 
 * DWARF framework under C++. Mainly conversion routines between different
 * representations of rotations (axis/angle, quaternions, rotation matrices,
								 * euler angles)
 *
 * @author Martin Wagner
 */ 

/*
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 * $Id: dwarfutil.h,v 1.1 2004/05/03 09:35:00 wagnerm Exp $
 * $Revision: 1.1 $
 */

#ifndef __MATH_LIBRARY_H_
#define __MATH_LIBRARY_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _WIN32
#include <windows.h>
#include <time.h>
#else
#include <sys/time.h>
#endif




//! Tolerance for quaternion operations
const double Q_EPSILON = (1e-10);
//! Tolerance for unit quaternion checking
const double Q_UNITCHECK = (1e-5);

//! macro to check whether quaternion is a unit quaternion
#define checkUnit(q) \
	assert( fabs(q[QX]*q[QX] + q[QY]*q[QY] + q[QZ]*q[QZ] + q[QW]*q[QW] - 1) <= Q_UNITCHECK)


/*
 * Some constants and methods for accessing quaternion components
 */
//! @brief Use only this constant for accessing a quaternion's X component
const unsigned char QX = 0;
//! @brief Use only this constant for accessing a quaternion's Y component
const unsigned char QY = 1;
//! @brief Use only this constant for accessing a quaternion's Z component
const unsigned char QZ = 2;
//! @brief Use only this constant for accessing a quaternion's scalar component
const unsigned char QW = 3;
//! @brief Convenience method for accessing a quaternion's X component
__inline double q_x( double *q) { return q[QX]; }
//! @brief Convenience method for accessing a quaternion's y component
__inline double q_y( double *q) { return q[QY]; }
//! @brief Convenience method for accessing a quaternion's z component
__inline double q_z( double *q) { return q[QZ]; }
//! @brief Convenience method for accessing a quaternion's scalar component
__inline double q_w( double *q) { return q[QW]; }


/**
 * @brief Get the current system time as a DWARF::Time struct
 *
 * @return: the current system time
 */

struct timeval currentTime();


/**
* @brief Construct rotation matrix from unit quaternion.
 *
 * Assumes matrix is used to multiply column vector on the left:
 * vnew = mat vold. Works correctly for right-handed coordinate system
 * and right-handed rotations. 
 *
 * @param quat 4-dim unit quaternion as input
 * @param matrix 4x4-dim double homogeneous rotation matrix
 *                as return value, organized as follows (row major)<br>
 *                <b>NOTE: OpenGL has another order, column major!</b><br>
 * <CODE>           ( 0  1  2  3 )<br>
 *                  ( 4  5  6  7 )<br>
 *                  ( 8  9 10 11 )<br>
 *                  (12 13 14 15 )
 * </CODE>
 * @return pointer to matrix
 */
double *quaternionToMatrix( const double *quat, double *matrix);

/**
* @brief Construct a unit quaternion from a rotation matrix. 
 * 
 * Assumes matrix is 
 * used to multiply column vector on the left: vnew = mat vold. Works 
 * correctly for right-handed coordinate system and right-handed rotations. 
 * Translation components are ignored. Matrix is assumed to be orthonormal.
 *
 * @param matrix 4x4-dim double homogeneous matrix
 *                as input, organized as follows (row major)<br>
 *                <b>NOTE: OpenGL has another order, column major!</b><br>
 * <CODE>           ( 0  1  2  3 )<br>
 *                  ( 4  5  6  7 )<br>
 *                  ( 8  9 10 11 )<br>
 *                  (12 13 14 15 )
 * </CODE>
 * @param quat 4-dim normalized quaternion as return value
 * @return pointer to quat
 */ 
double *matrixToQuaternion( const double *matrix, double *quat );


/**
* @brief Converts axis/angle representation (as used in VRML) to a unit quaternion
 * 
 * @param axisAngle 4-dim double axis/angle rotation (x,y,z,alpha) as input
 * @param quat 4-dim unit quaternion as return value
 * @return a pointer to quat
 */
double* axisAngleToQuaternion( double* axisAngle, double* quat );

/**
* @brief Converts a unit quaternion to axis/angle representation (as used in VRML)
 *
 * @param quat 4-dim unit quaternion as input
 * @param axisAngle 4-dim double axis/angle rotation (x,y,z,alpha) 
 *              as return value
 * @return a pointer to axisAngle
 */
double* quaternionToAxisAngle( double* quat, double* axisAngle );

/**
* @brief Inverts a quaternion. 
 *
 * The resulting quaternion is the inverse of the given quaternion. This
 * corresponds to translating an orthonormal rotation matrix.
 * Note: To invert a unit quaternion, conjugating it suffices.
 *
 * @param quat 4-dim quaternion as input and return value
 * @return a pointer to quat
 */
double* invertQuaternion( double *quat );

/**
* @brief Conjugates a quaternion (i.e. negates its vector part)
 * 
 * Note: Conjugating a unit quaternion results in its inverse.
 *
 * @param quat 4-dim quaternion as input and return value
 * @return a pointer to quat
 */
double* conjugateQuaternion( double *quat );

/**
* @brief Normalizes a quaternion (i.e. makes a unit quaternion)
 * 
 * @param quat 4-dim quaternion as input and return value
 * @return a pointer to quat
 */
double* normalizeQuaternion( double *quat );

/**
* multiplyQuaternion
 * Calculates <CODE> quat := quat1 * quat2 </CODE>
 *
 * @param quat 4-dim Quaternion as return value
 * @param quat1 4-dim Quaternion as input
 * @param quat2 4-dim Quaternion as input 
 * @return a pointer to quat
 */
double* multiplyQuaternion( double *quat, const double *quat1, const double *quat2 );

/**
* @brief Rotates a 3-dim point by a unit quaternion
 *
 * Calculates <CODE> rotPoint = quat * (point,0) * quat^-1 </CODE>
 *
 * @param quat 4-dim unit quaternion representing the rotation
 * @param point 3-dim point to be rotated as input
 * @param rotPoint memory for return value (at least 3*double)
 * @return a pointer to rotPoint
 */
double* rotateQuaternion( double *rotPoint, const double *quat, const double *point );

/**
* @brief Rotates a 3-dim point by a orthonormal rotation matrix
 *
 * Calculates <CODE> rotPoint = matrix * (point,1) </CODE>
 *
 * @param quat 4x4-dim row major orthonormal matrix representing the rotation, will not be changed
 * @param point 3-dim point to be rotated as input
 * @param rotPoint memory for return value (at least 3*double)
 * @return a pointer to rotPoint
 */
double* rotateMatrix( double *rotPoint, double *matrix, const double *point );


/**
* @brief Inverts a pose transformation.
 *
 * Identical to inverting a homogeneous 4x4 matrix, computes
 * rot_new = rot_old^-1; t_new = - rot_new * t_old
 *
 * @param newQuat 4-dim unit quaternion, new rotation (return value)
 * @param newTrans 3-dim pont, new translation (return value)
 * @param oldQuat 4-dim unit quaternion, old rotation
 * @param oldTrans 3-dim point, old translation
 */
void invertQuaternionPose( double* newQuat, double* newTrans, const double* oldQuat, const double* oldTrans );


/**
 * @brief makes sure a quaternion is close to another given quaternion
 *
 * Unit Quaternions can be regarded as points on a 4-dim hypersphere.
 * Two opposing points on this sphere express the same rotation.
 * this function chooses the point nearer to a reference quaternion.
 *
 * @param quat 4-dim unit quaternion, maybe negated at the end
 * @param otherquat reference 4-dim unit quaternion
 */
double* quaternionNegateIfCloser( double* quat, const double* otherquat );

#endif

#ifdef __cplusplus
}
#endif
